@@ -11,11 +11,19 @@ import copy
 from os import path
 from dataclasses import dataclass
 from enum import Enum
+import re
+
+
+class StockfishException(Exception):
+    pass
 
 
 class Stockfish:
     """Integrates the Stockfish chess engine with Python."""
 
+    _del_counter = 0
+    # Used in test_models: will count how many times the del function is called.
+
     def __init__(
         self, path: str = "stockfish", depth: int = 15, parameters: dict = None
     ) -> None:
@@ -25,7 +33,7 @@ class Stockfish:
             "Min Split Depth": 0,
             "Threads": 1,
             "Ponder": "false",
-            "Hash": 1024,
+            "Hash": 16,
             "MultiPV": 1,
             "Skill Level": 20,
             "Move Overhead": 10,
@@ -35,8 +43,9 @@ class Stockfish:
             "UCI_LimitStrength": "false",
             "UCI_Elo": 1350,
         }
+        self._path = path
         self._stockfish = subprocess.Popen(
-            path,
+            self._path,
             universal_newlines=True,
             stdin=subprocess.PIPE,
             stdout=subprocess.PIPE,
@@ -46,7 +55,7 @@ class Stockfish:
         self._has_quit_command_been_sent = False
 
         self._stockfish_major_version: int = int(
-            self._read_line().split(" ")[1].split(".")[0]
+            self._read_line().split(" ")[1].split(".")[0].replace("-", "")
         )
 
         self._put("uci")
@@ -147,6 +156,8 @@ class Stockfish:
     def _read_line(self) -> str:
         if not self._stockfish.stdout:
             raise BrokenPipeError()
+        if self._stockfish.poll() is not None:
+            raise StockfishException("The Stockfish process has crashed")
         return self._stockfish.stdout.readline().strip()
 
     def _set_option(
@@ -159,9 +170,8 @@ class Stockfish:
 
     def _is_ready(self) -> None:
         self._put("isready")
-        while True:
-            if self._read_line() == "readyok":
-                return
+        while self._read_line() != "readyok":
+            pass
 
     def _go(self) -> None:
         self._put(f"go depth {self.depth}")
@@ -177,14 +187,27 @@ class Stockfish:
             cmd += f" btime {btime}"
         self._put(cmd)
 
-    @staticmethod
-    def _convert_move_list_to_str(moves: List[str]) -> str:
-        result = ""
-        for move in moves:
-            result += f"{move} "
-        return result.strip()
+    def set_fen_position(
+        self, fen_position: str, send_ucinewgame_token: bool = True
+    ) -> None:
+        """Sets current board position in Forsyth–Edwards notation (FEN).
 
-    def set_position(self, moves: List[str] = None) -> None:
+        Args:
+            fen_position:
+              FEN string of board position.
+
+            send_ucinewgame_token:
+              Whether to send the "ucinewgame" token to the Stockfish engine.
+              The most prominent effect this will have is clearing Stockfish's transposition table,
+              which should be done if the new position is unrelated to the current position.
+
+        Returns:
+            None
+        """
+        self._prepare_for_new_position(send_ucinewgame_token)
+        self._put(f"position fen {fen_position}")
+
+    def set_position(self, moves: Optional[List[str]] = None) -> None:
         """Sets current board position.
 
         Args:
@@ -193,12 +216,12 @@ class Stockfish:
               Must be in full algebraic notation.
               example: ['e2e4', 'e7e5']
         """
-        self._prepare_for_new_position(True)
-        if moves is None:
-            moves = []
-        self._put(f"position startpos moves {self._convert_move_list_to_str(moves)}")
+        self.set_fen_position(
+            "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1", True
+        )
+        self.make_moves_from_current_position(moves)
 
-    def make_moves_from_current_position(self, moves: List[str]) -> None:
+    def make_moves_from_current_position(self, moves: Optional[List[str]]) -> None:
         """Sets a new position by playing the moves from the current position.
 
         Args:
@@ -207,32 +230,56 @@ class Stockfish:
               Must be in full algebraic notation.
               Example: ["g4d7", "a8b8", "f1d1"]
         """
-        if moves == []:
-            raise ValueError(
-                "No moves sent in to the make_moves_from_current_position function."
-            )
+        if not moves:
+            return
         self._prepare_for_new_position(False)
-        self._put(
-            f"position fen {self.get_fen_position()} moves {self._convert_move_list_to_str(moves)}"
-        )
+        for move in moves:
+            if not self.is_move_correct(move):
+                raise ValueError(f"Cannot make move: {move}")
+            self._put(f"position fen {self.get_fen_position()} moves {move}")
 
-    def get_board_visual(self) -> str:
+    def get_board_visual(self, perspective_white: bool = True) -> str:
         """Returns a visual representation of the current board position.
 
+        Args:
+            perspective_white:
+              A bool that indicates whether the board should be displayed from the
+              perspective of white (True: white, False: black)
+
         Returns:
             String of visual representation of the chessboard with its pieces in current position.
         """
         self._put("d")
-        board_rep = ""
+        board_rep_lines = []
         count_lines = 0
         while count_lines < 17:
             board_str = self._read_line()
             if "+" in board_str or "|" in board_str:
                 count_lines += 1
-                board_rep += f"{board_str}\n"
-        if self._stockfish_major_version >= 12:
-            board_str = self._read_line()
-            board_rep += f"  {board_str}\n"
+                if perspective_white:
+                    board_rep_lines.append(f"{board_str}")
+                else:
+                    # If the board is to be shown from black's point of view, all lines are
+                    # inverted horizontally and at the end the order of the lines is reversed.
+                    board_part = board_str[:33]
+                    # To keep the displayed numbers on the right side,
+                    # only the string representing the board is flipped.
+                    number_part = board_str[33:] if len(board_str) > 33 else ""
+                    board_rep_lines.append(f"{board_part[::-1]}{number_part}")
+        if not perspective_white:
+            board_rep_lines = board_rep_lines[::-1]
+        board_str = self._read_line()
+        if "a   b   c" in board_str:
+            # Engine being used is recent enough to have coordinates, so add them:
+            if perspective_white:
+                board_rep_lines.append(f"  {board_str}")
+            else:
+                board_rep_lines.append(f"  {board_str[::-1]}")
+        while "Checkers" not in self._read_line():
+            # Gets rid of the remaining lines in _stockfish.stdout.
+            # "Checkers" is in the last line outputted by Stockfish for the "d" command.
+            pass
+        board_rep = "\n".join(board_rep_lines) + "\n"
         return board_rep
 
     def get_fen_position(self) -> str:
@@ -246,6 +293,8 @@ class Stockfish:
             text = self._read_line()
             splitted_text = text.split(" ")
             if splitted_text[0] == "Fen:":
+                while "Checkers" not in self._read_line():
+                    pass
                 return " ".join(splitted_text[1:])
 
     def set_skill_level(self, skill_level: int = 20) -> None:
@@ -275,26 +324,6 @@ class Stockfish:
             {"UCI_LimitStrength": "true", "UCI_Elo": elo_rating}
         )
 
-    def set_fen_position(
-        self, fen_position: str, send_ucinewgame_token: bool = True
-    ) -> None:
-        """Sets current board position in Forsyth–Edwards notation (FEN).
-
-        Args:
-            fen_position:
-              FEN string of board position.
-
-            send_ucinewgame_token:
-              Whether to send the "ucinewgame" token to the Stockfish engine.
-              The most prominent effect this will have is clearing Stockfish's transposition table,
-              which should be done if the new position is unrelated to the current position.
-
-        Returns:
-            None
-        """
-        self._prepare_for_new_position(send_ucinewgame_token)
-        self._put(f"position fen {fen_position}")
-
     def get_best_move(self, wtime: int = None, btime: int = None) -> Optional[str]:
         """Returns best move with current position on the board.
         wtime and btime arguments influence the search only if provided.
@@ -306,16 +335,7 @@ class Stockfish:
             self._go_remaining_time(wtime, btime)
         else:
             self._go()
-        last_text: str = ""
-        while True:
-            text = self._read_line()
-            splitted_text = text.split(" ")
-            if splitted_text[0] == "bestmove":
-                if splitted_text[1] == "(none)":
-                    return None
-                self.info = last_text
-                return splitted_text[1]
-            last_text = text
+        return self._get_best_move_from_sf_popen_process()
 
     def get_best_move_time(self, time: int = 1000) -> Optional[str]:
         """Returns best move with current position on the board after a determined time
@@ -328,17 +348,75 @@ class Stockfish:
             A string of move in algebraic notation or None, if it's a mate now.
         """
         self._go_time(time)
+        return self._get_best_move_from_sf_popen_process()
+
+    def _get_best_move_from_sf_popen_process(self) -> Optional[str]:
+        # Precondition - a "go" command must have been sent to SF before calling this function.
+        # This function needs existing output to read from the SF popen process.
         last_text: str = ""
         while True:
             text = self._read_line()
             splitted_text = text.split(" ")
             if splitted_text[0] == "bestmove":
-                if splitted_text[1] == "(none)":
-                    return None
                 self.info = last_text
-                return splitted_text[1]
+                return None if splitted_text[1] == "(none)" else splitted_text[1]
             last_text = text
 
+    @staticmethod
+    def _is_fen_syntax_valid(fen: str) -> bool:
+        # Code for this function taken from: https://gist.github.com/Dani4kor/e1e8b439115878f8c6dcf127a4ed5d3e
+        # Some small changes have been made to the code.
+
+        regexMatch = re.match(
+            r"\s*^(((?:[rnbqkpRNBQKP1-8]+\/){7})[rnbqkpRNBQKP1-8]+)\s([b|w])\s(-|[K|Q|k|q]{1,4})\s(-|[a-h][1-8])\s(\d+\s\d+)$",
+            fen,
+        )
+        if not regexMatch:
+            return False
+        regexList = regexMatch.groups()
+        if len(regexList[0].split("/")) != 8:
+            return False  # 8 rows not present.
+        for fenPart in regexList[0].split("/"):
+            field_sum = 0
+            previous_was_digit = False
+            for c in fenPart:
+                if c in ["1", "2", "3", "4", "5", "6", "7", "8"]:
+                    if previous_was_digit:
+                        return False  # Two digits next to each other.
+                    field_sum += int(c)
+                    previous_was_digit = True
+                elif c.lower() in ["p", "n", "b", "r", "q", "k"]:
+                    field_sum += 1
+                    previous_was_digit = False
+                else:
+                    return False  # Invalid character.
+            if field_sum != 8:
+                return False  # One of the rows doesn't have 8 columns.
+        return True
+
+    def is_fen_valid(self, fen: str) -> bool:
+        if not Stockfish._is_fen_syntax_valid(fen):
+            return False
+        temp_sf = Stockfish(path=self._path, parameters={"Hash": 1})
+        # Using a new temporary SF instance, in case the fen is an illegal position that causes
+        # the SF process to crash.
+        best_move = None
+        temp_sf.set_fen_position(fen, False)
+        try:
+            temp_sf._put("go depth 10")
+            best_move = temp_sf._get_best_move_from_sf_popen_process()
+        except StockfishException:
+            # If a StockfishException is thrown, then it happened in read_line() since the SF process crashed.
+            # This is likely due to the position being illegal, so set the var to false:
+            return False
+        else:
+            return best_move is not None
+        finally:
+            temp_sf.__del__()
+            # Calling this function before returning from either the except or else block above.
+            # The __del__ function should generally be called implicitly by python when this
+            # temp_sf object goes out of scope, but calling it explicitly guarantees this will happen.
+
     def is_move_correct(self, move_value: str) -> bool:
         """Checks new move.
 
@@ -349,12 +427,11 @@ class Stockfish:
         Returns:
             True, if new move is correct, else False.
         """
+        old_self_info = self.info
         self._put(f"go depth 1 searchmoves {move_value}")
-        while True:
-            text = self._read_line()
-            splitted_text = text.split(" ")
-            if splitted_text[0] == "bestmove":
-                return splitted_text[1] != "(none)"
+        is_move_correct = self._get_best_move_from_sf_popen_process() is not None
+        self.info = old_self_info
+        return is_move_correct
 
     def get_wdl_stats(self) -> Optional[List]:
         """Returns Stockfish's win/draw/loss stats for the side to move.
@@ -398,13 +475,17 @@ class Stockfish:
         """
 
         self._put("uci")
+        encountered_UCI_ShowWDL = False
         while True:
             text = self._read_line()
             splitted_text = text.split(" ")
             if splitted_text[0] == "uciok":
-                return False
+                return encountered_UCI_ShowWDL
             elif "UCI_ShowWDL" in splitted_text:
-                return True
+                encountered_UCI_ShowWDL = True
+                # Not returning right away, since the remaining lines should be read and
+                # discarded. So continue the loop until reaching "uciok", which is
+                # the last line SF outputs for the "uci" command.
 
     def get_evaluation(self) -> dict:
         """Evaluates current position
@@ -415,10 +496,9 @@ class Stockfish:
 
         evaluation = dict()
         fen_position = self.get_fen_position()
-        if "w" in fen_position:  # w can only be in FEN if it is whites move
-            compare = 1
-        else:  # stockfish shows advantage relative to current player, convention is to do white positive
-            compare = -1
+        compare = 1 if "w" in fen_position else -1
+        # Stockfish shows advantage relative to current player. This function will instead
+        # use positive to represent advantage white, and negative for advantage black.
         self._put(f"position {fen_position}")
         self._go()
         while True:
@@ -543,14 +623,11 @@ class Stockfish:
         self._put(
             f"bench {params.ttSize} {params.threads} {params.limit} {params.fenFile} {params.limitType} {params.evalType}"
         )
-        last_text: str = ""
         while True:
             text = self._read_line()
             splitted_text = text.split(" ")
             if splitted_text[0] == "Nodes/second":
-                last_text = text
-                return last_text
-            last_text = text
+                return text
 
     def set_depth(self, depth_value: int = 2) -> None:
         """Sets current depth of stockfish engine.
@@ -627,7 +704,7 @@ class Stockfish:
         if not self.is_move_correct(move_value):
             raise ValueError("The proposed move is not valid in the current position.")
         starting_square_piece = self.get_what_is_on_square(move_value[:2])
-        ending_square_piece = self.get_what_is_on_square(move_value[-2:])
+        ending_square_piece = self.get_what_is_on_square(move_value[2:4])
         if ending_square_piece != None:
             if self._parameters["UCI_Chess960"] == "false":
                 return Stockfish.Capture.DIRECT_CAPTURE
@@ -641,7 +718,7 @@ class Stockfish:
                     return Stockfish.Capture.NO_CAPTURE
                 else:
                     return Stockfish.Capture.DIRECT_CAPTURE
-        elif move_value[-2:] == self.get_fen_position().split()[
+        elif move_value[2:4] == self.get_fen_position().split()[
             3
         ] and starting_square_piece in [
             Stockfish.Piece.WHITE_PAWN,
@@ -676,8 +753,8 @@ class Stockfish:
         )
 
     def __del__(self) -> None:
+        Stockfish._del_counter += 1
         if self._stockfish.poll() is None:
             self._put("quit")
-            self._stockfish.kill()
             while self._stockfish.poll() is None:
                 pass
